# Copyright The IETF Trust 2018, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division

import collections
import io
import jinja2
import lxml
import os
import pydoc
import re
import yaml

from xml2rfc import debug, __version__
from xml2rfc.utils import namespaces
from xml2rfc.writers import base
from xml2rfc import util

debug = debug                           # silence pyflakes


def capfirst(value):
    """Capitalize the first character of the value."""
    return value and value[0].upper() + value[1:]

class DocWriter(base.BaseV3Writer):
    def __init__(self, xmlrfc, quiet=None, options=base.default_options, date=None):
        super(DocWriter, self).__init__(xmlrfc, quiet=quiet, options=options, date=date)
        rfc7991_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'rfc7991.rng')
        self.rfc7991_schema = lxml.etree.ElementTree(file=rfc7991_rng_file)
        self.template_dirs = list(filter(None, [ options.template_dir, os.path.join(os.path.dirname(os.path.dirname(__file__)), 'templates') ]))
        self.jinja = jinja2.Environment(
            loader = jinja2.FileSystemLoader(self.template_dirs),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.jinja.filters['capfirst'] = capfirst
        self.rendered = None

    def get_elements(self):
        elements = {}
        parents = collections.defaultdict(set)
        attribtag = '{%s}attribute'%namespaces['x']
        ignored_attributes = [

            'xml:base',
            'xml:lang',
            'xml:space',

            'derivedAnchor',
            'derivedContent',
            'derivedCounter',
            'derivedLink',
            'displayFormat',
            'expiresDate',
            'mode',
            'originalSrc',
            'pn',
            'quoteTitle',
            'pageno',
            'scripts',
            'slugifiedName',

        ]

        #debug.show('attribtag')
        edefs = self.v3_schema.xpath("/x:grammar/x:define/x:element", namespaces=namespaces)
        rnc = RelaxNGCompactRenderer()
        rnc.render(self.v3_schema.getroot())
        for element in edefs:
            ename = element.get('name')
            is_new = self.rfc7991_schema.xpath("/x:grammar/x:define/x:element[@name='%s']" % ename, namespaces=namespaces) == []
            e = {'attributes': [], 'children': [], 'rnc': [], 'parents': [], 'new': is_new, }
            elements[ename] = e
            #debug.show('ename')
            for a in element.xpath('.//x:attribute', namespaces=namespaces):
                aname = a.get('name')
                if aname in ignored_attributes:
                    continue
                adict = {'name': aname}
                e['attributes'].append(adict)
                adict['required'] = a.getparent().tag != '{http://relaxng.org/ns/structure/1.0}optional'
                default = a.get('{http://relaxng.org/ns/compatibility/annotations/1.0}defaultValue')
                if default:
                    adict['default'] = default
                choices = a.xpath('x:choice/x:value', namespaces=namespaces)
                if choices:
                    adict['choices'] = [ dict(value=v.text) for v in choices ]
                    adict['rnc'] = rnc.render(a.getchildren()[0])
                adict['new'] = self.rfc7991_schema.xpath("/x:grammar/x:define/x:element[@name='%s']//x:attribute[@name='%s']" % (ename, aname), namespaces=namespaces) == []
                adict['deprecated'] = (ename, aname) in base.deprecated_attributes
                #debug.show("e['attributes']")
            for c in element.getchildren():
                if c.tag != attribtag and c.find(attribtag) == None:
                    e['children'].append((c, rnc.render(c)))
                    cnames = element.xpath('.//x:ref/@name', namespaces=namespaces)
                    for cname in cnames:
                        parents[cname].add(ename)
            e['attributes'].sort(key=lambda x: x['name'])
            e['deprecated_attributes'] = any(a['deprecated'] for a in e['attributes'])
            e['rnc'] = ', '.join(filter(None, [ c[1] for c in e['children'] ]))
        for k, v in parents.items():
            if k in elements:
                elements[k]['parents'] = list(v)
                elements[k]['parents'].sort()
            
        return elements


    def process(self):
        from xml2rfc.run import optionparser
        if self.rendered:
            return self.rendered

        template = self.jinja.get_template('doc.xml')

        # --- Element and attribute information for the template context ---
        elements = self.get_elements()

        for tdir in self.template_dirs:
            fn = os.path.join(tdir, 'doc.yaml')
            if os.path.exists(fn):
                with io.open(fn) as file:
                    text = file.read()
                    text = text.replace(r'\<', r'&lt;').replace(r'\>', r'&gt;').replace(r'\&', r'&amp;')
                    descriptions = yaml.load(text, Loader=yaml.FullLoader)

        for n, d in elements.items():
            d['description'] = descriptions.get(n, None)
            for dd in d['attributes']:
                a = dd['name']
                dd['description'] = descriptions.get('%s[%s]' % (n,a), None)
                choices = dd.get('choices', [])
                for ddd in choices:
                    c = ddd['value']
                    ddd['description'] = descriptions.get('%s[%s="%s"]' % (n,a, c), None)
                dd['choice_descriptions'] = any(ddd['description'] for ddd in choices)
        element_list = [ {'tag': t, 'deprecated': t in base.deprecated_element_tags } for t in base.element_tags ]
        element_list.sort(key=lambda x: x['tag'])

        for d in element_list:
            d.update(elements[d['tag']])

        # --- Command-line options information for the template context ---

        # Here's a silly little dance because configargparse puts its 'is_config_file_arg=True'
        # arguments in 'optional arguments', even if we've explicitly put it in another group
        for a in optionparser._actions:
            # This depend on at least one generic option with title being listed before '--config-file'
            if a.container.title == 'Generic Options with Values':
                genargs_group = a.container
            if a.container.title == 'optional arguments' or a.container.title == 'options':
                # Python 3.10 calls this 'options' while Python 3.6 to Python 3.9 calls this 'optional arguments'
                a.container = genargs_group
        # Deal with options that has an inverse form
        option_strings = { a.option_strings[-1]: a for a in optionparser._actions if a.option_strings }
        for o, a in option_strings.items():
            a.suppress = "==SUPPRESS==" in a.help
        option_strings = { k: a for (k, a) in option_strings.items() if not a.suppress }
        for o, a in option_strings.items():
            try:
                a.has_negation = o.replace('--', '--no-') in option_strings 
                a.has_positive = '--no-' in o and o.replace('--no-', '--') in option_strings
            except TypeError:
                pass

        # Provide a count of options in each group, for the template
        for i, group in enumerate(optionparser._action_groups):
            group.options = len([ o for o in group._actions if o.container==group ])

        # --- Context variable information for the template context ---
        descriptions['-context-'] = {}
        descriptions['-context-']['descriptions']    = "The descriptions read from the <tt>doc.yaml</tt> file"
        descriptions['-context-']['element_tags']    = "A list of all the element tags in the XML schema"
        descriptions['-context-']['elements']        = ("A list of dictionaries, each with information about one schema element: "
            '{'+''.join( '"%s": ..., '% k for k in sorted(element_list[0].keys())) +'}')
        descriptions['-context-']['options']         = "A list of dictionaries describing the command-line options"
        descriptions['-context-']['schema']          = "The full RelaxNG Compact representation of the schema in text form"
        descriptions['-context-']['toc_depth']       = "ToC depth setting; 1 when running --man, 2 otherwise"
        descriptions['-context-']['v3_element_tags'] = "A list of v3 element tags, excluding deprecated tags"
        descriptions['-context-']['version']         = "The xml2rfc version number"


        # --- Set up context ---
        context = {}
        context['elements'] = element_list
        context['element_tags'] = list(base.element_tags)
        context['options'] = optionparser
        context['toc_depth'] = getattr(self.options, 'toc_depth', "2")
        context['v3_element_tags'] = list(base.element_tags-base.deprecated_element_tags)
        context['bare_latin_tags'] = util.unicode.bare_latin_tags
        context['version'] =  __version__
        context['descriptions'] = descriptions
        with open(self.v3_rnc_file) as file:
            context['schema'] = ('\n' + file.read()).replace('\n   ', '\n')

        # Context meta-information
        context['context'] = context

        # Template rendering, using context
        self.rendered = template.render(context)

        return self.rendered

    def manpage(self):
        import xml2rfc
        parser = xml2rfc.XmlRfcParser(None, options=self.options)
        self.options.toc_depth = "1"
        text = self.process()
        parser.text = text.encode('utf8')
        xmlrfc = parser.parse(remove_comments=False, quiet=True, add_xmlns=True)
        preptool = xml2rfc.PrepToolWriter(xmlrfc, options=self.options, date=self.options.date, liberal=True, keep_pis=[xml2rfc.V3_PI_TARGET])
        xmlrfc.tree = preptool.prep()
        if xmlrfc.tree:
            self.options.pagination = False
            writer = xml2rfc.TextWriter(xmlrfc, options=self.options, date=self.options.date)
            text = writer.process()

        if writer.errors:
            raise base.RfcWriterError("Cannot show manpage due to errors.")
        else:
            pydoc.pager(text.lstrip())

    def write(self, filename):
        self.process()

        with io.open(filename, "w") as file:
            file.write(self.rendered)

        if not self.options.quiet:
            self.log(' Created file %s' % filename)


class RelaxNGCompactRenderer():
    "Simple (and incomplete) converter from an RNG tree to RNC text; enough for our purpose."
    #
    keywords = ['attribute', 'default', 'datatypes', 'div', 'element', 'empty', 'external',
                'grammar', 'include', 'inherit', 'list', 'mixed', 'namespace', 'notAllowed',
                'parent', 'start', 'string', 'text', 'token', ]
    deprecated = base.deprecated_element_tags

    def __init__(self, **kwargs):
        self.nsnames = {}
        if 'deprecated' in kwargs:
            self.deprecated = kwargs.pop('deprecated')
    def render(self, e):
        if e.tag  in [lxml.etree.Comment, ]:
            return ''
        tag = e.tag.replace('{http://relaxng.org/ns/structure/1.0}', '')
        if hasattr(self, tag):
            func = getattr(self, tag)
            return func(e)
        else:
            raise NotImplementedError('%s()' % tag)
    def zeroOrMore(self, e):
        cc = e.getchildren()
        sub = ', '.join(filter(None, [ self.render(c) for c in cc ]))
        sub = '( %s )' % sub if len(cc) > 1 else sub
        return sub + ' *' if sub else ''
    def oneOrMore(self, e):
        cc = e.getchildren()
        sub = ', '.join(filter(None, [ self.render(c) for c in cc ]))
        sub = '( %s )' % sub if len(cc) > 1 else sub
        return sub + ' +' if sub else ''
    def ref(self, e):
        txt = e.get('name')
        if txt in self.deprecated:
            return ''
        if txt in self.keywords:
            txt = r'\%s' % txt
        return txt
    def optional(self, e):
        cc = e.getchildren()
        sub = ', '.join(filter(None, [ self.render(c) for c in cc ]))
        sub = '(%s)' % sub if len(cc) > 1 else sub
        return sub + ' ?' if sub else ''
    def choice(self, e):
        cc = e.getchildren()
        sub = ' | '.join(filter(None, [ self.render(c) for c in cc ]))
        sub = '( %s )' % sub if len(cc) > 1 else sub
        sub = sub.replace(')* | (', ')*\n| (')
        sub = sub.replace(') | (', ')\n| (')
        sub = sub.replace(')+ | (', ')+\n| (')
        return sub
    def interleave(self, e):
        cc = e.getchildren()
        sub = ' & '.join(filter(None, [ self.render(c) for c in cc ]))
        sub = '( %s )' % sub if len(cc) > 1 else sub
        sub = sub.replace(')* & (', ')*\n& (')
        sub = sub.replace(') & (', ')\n& (')
        sub = sub.replace(')+ & (', ')+\n& (')
    def text(self, e):
        return 'text'
    def empty(self, e):
        return 'empty'
    def value(self, e):
        return '"%s"' % e.text
    def grammar(self, e):
        text = '\n'
        self.nsmap = e.nsmap
        for a in e.nsmap.keys():
            v = e.nsmap.get(a)
            self.nsnames[v] = a
            if a:
                text += 'namespace %s = "%s"\n' % (a, v)
        for c in e.getchildren():
            text += self.render(c) + '\n'
        return text
    def define(self, e):
        txt = e.get('name')
        if txt in self.keywords:
            txt = r'\%s' % txt
        cc = e.getchildren()
        sub = ',\n'.join(filter(None, [ self.render(c) for c in cc ]))
        return '%s =\n  %s\n' % (txt, sub)
    def element(self, e):
        name = e.get('name')
        cc = e.getchildren()
        sub = ',\n    '.join(filter(None, [ self.render(c) for c in cc ]))
        txt = 'element %s {\n    %s\n  }\n' % (name, sub)
        txt = re.sub(r'(\n[ \t]*)+\n', '\n', txt)
        return txt
    def attribute(self, e):
        txt = ''
        name = e.get('name')
        nsa = 'http://relaxng.org/ns/compatibility/annotations/1.0'
        default = e.get('{%s}defaultValue' % nsa, None)
        cc = e.getchildren()
        sub = ', '.join(filter(None, [ self.render(c) for c in cc ])) if cc else 'text'
        if len(cc)==1 and re.match(r'\(.*\)$', sub):
            sub = sub[1:-1]
        if default != None:
            txt += '\n    [ %s:defaultValue = "%s" ]\n    ' % (self.nsnames[nsa], default)
        txt += 'attribute %s { %s }' % (name, sub)
        return txt
    def data(self, e):
        type = e.get('type')
        return 'xsd:%s' % type
    def include(self, e):
        return 'include "%s"' % e.get('href').replace('.rng', '.rnc')
    def start(self, e):
        combine = e.get('combine', None)
        if combine == None:
            asg = '='
        elif combine == 'choice':
            asg = '|='
        elif combine == 'interleave':
            asg = '&='
        else:
            raise NotImplementedError('%s' % e)
        cc = e.getchildren()
        sub = ', '.join(filter(None, [ self.render(c) for c in cc ]))
        return 'start %s %s' % (asg, sub)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright The IETF Trust 2018, All Rights Reserved

from __future__ import unicode_literals, print_function, division

import dict2xml
import io
import json
import lxml
import os
import pypdf
import sys

def walk(obj, seen):
    dobj = {}                            # Direct objects
    iobj = []                            # Indirect objects
    if isinstance(obj, pypdf.generic.DictionaryObject):
        for key in obj.keys():
            k = key[1:] if key.startswith('/') else key
            d, i = walk(obj[key], seen)
            dobj[k] = d
            iobj += i
        if hasattr(obj, 'extract_text'):
            dobj['text'] = obj.extract_text(extraction_mode="layout")
    elif isinstance(obj, pypdf.generic.ArrayObject):
        dobj = []
        for o in obj:
            d, i = walk(o, seen)
            dobj.append(d)
            iobj += i
    elif isinstance(obj, pypdf.generic.BooleanObject):
        dobj = obj.value
    elif isinstance(obj, pypdf.generic.NameObject):
        dobj = str(obj)
    elif isinstance(obj, pypdf.generic.NumberObject):
        dobj = int(obj)
    elif isinstance(obj, pypdf.generic.FloatObject):
        dobj = float(obj)
    elif isinstance(obj, pypdf.generic.IndirectObject):
        dobj = str(obj)
        if (obj.idnum, obj.generation) not in seen:
            seen.add((obj.idnum, obj.generation))
            d, i = walk(obj.get_object(), seen)
            if isinstance(d, dict):
                d['IdNum'] = obj.idnum
                d['Generation'] = obj.generation
            else:
                dobj = d
            iobj += i
            iobj.append(d)
    elif isinstance(obj, pypdf.generic.TextStringObject):
        dobj = str(obj)
    else:
        raise RuntimeError("Unexpected object type: %s" % type(obj))

    if hasattr(obj, 'idnum'):
        seen.add((obj.idnum, obj.generation))

    return dobj, iobj

def pyobj(filename=None, bytes=None):
    seen = set()
    #
    pdffile = io.BytesIO(bytes) if bytes else io.open(filename, 'br')
    reader = pypdf.PdfReader(pdffile, strict=False)
    info = reader.metadata
    doc = {}
    for key in info.keys():
        k = key[1:] if key.startswith('/') else key
        doc[k] = info[key]
    iobj = []
    pages = []
    for num in range(len(reader.pages)):
        page = reader.pages[num]
        obj = page.get_object()
        d, i = walk(obj, seen)
        #pages[num+1] = d
        pages.append(d)
        iobj += i
    pdffile.close()
    #
    doc['Page'] = pages
    doc['IndirectObject'] = iobj
    return doc

def xmltext(filename=None, obj=None, bytes=None):
    if obj is None:
        obj = pyobj(filename=filename, bytes=bytes)
#     for i,p in enumerate(obj['Pages']):
#         obj['Pages'][i] = {'Page': p}
    return dict2xml.dict2xml(obj, wrap="Document")    

def xmldoc(filename, text=None, bytes=None):
    if text is None:
        text = xmltext(filename=filename, bytes=bytes)
    return lxml.etree.fromstring(text)

def main():
    for filename in sys.argv[1:]:
        if not os.path.exists(filename):
            print('Could not find "%s"' % filename)
        print('File: %s' % filename)
        doc = pyobj(filename)
        with io.open(filename+'.json', 'w', encoding='utf-8') as j:
            json.dump(doc, j, indent=2)
        print('Wrote: %s' % j.name)
        with io.open(filename+'.xml', 'w', encoding='utf-8') as x:
            x.write(xmltext(filename, doc))
        print('Wrote: %s' % x.name)

if __name__ == "__main__":
    main()



# Copyright The IETF Trust 2018, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division

import math

# ----------------------------------------------------------------------
# Base conversions.
# From http://tech.vaultize.com/2011/08/python-patterns-number-to-base-x-and-the-other-way/

DEFAULT_DIGITS = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
 
def num_to_baseX(num, digits=DEFAULT_DIGITS):
   if num < 0: return '-' + num_to_baseX(-num)
   if num == 0: return digits[0]
   X = len(digits)
   s = ''
   while num > 0:
        s = digits[num % X] + s
        num //= X
 
   return s
 
def baseX_to_num(s, digits=DEFAULT_DIGITS):
   if s[0] == '-': return -1 * baseX_to_num(s[1:])
   ctopos = dict([(c, pos) for pos, c in enumerate(digits)])
   X = len(digits)
   num = 0
   for c in s: num = num * X + ctopos[c]
   return num


# Use the generic base conversion to create list letters

def int2letter(num):
    return num_to_baseX(num-1, "abcdefghijklmnopqrstuvwxyz")

def int2roman(number):
    numerals = { 
        1 : "i", 
        4 : "iv", 
        5 : "v", 
        9 : "ix", 
        10 : "x", 
        40 : "xl", 
        50 : "l", 
        90 : "xc", 
        100 : "c", 
        400 : "cd", 
        500 : "d", 
        900 : "cm", 
        1000 : "m" 
    }
    if number > 3999:
        raise NotImplementedError("Can't handle roman numerals larger than 3999")
    result = ""
    for value, numeral in sorted(numerals.items(), reverse=True):
        while number >= value:
            result += numeral
            number -= value
    return result


roman_max_widths = { 1:1,  2:2,  3:3,  4:3,  5:3,  6:3,  7:3,  8:4,  9:4,
                10:4, 11:4, 12:4, 13:4, 14:4, 15:4, 16:4, 17:4, 18:5, 19:5,
                20:5, 21:5, 22:5, 23:5, 24:5, 25:5, 26:5, 27:5, 28:6, 29:6, }

def update_roman_max_widths(n):
    if n > 3999:
        raise NotImplementedError("Can't handle roman numerals larger than 3999")
    m = len(roman_max_widths)
    wmax = 0
    for i in range(n+32):
        w = len(int2roman(i))
        if w > wmax:
            wmax = w
        if n > m:
            roman_max_widths[n] = wmax

def num_width(type, num):
    """
    Return the largest width taken by the numbering of a list
    with num items (without punctuation)
    """
    if   type in ['a','A','c','C',]:
        return int(math.log(num, 26))+1
    elif type in ['1','d',]:
        return int(math.log(num, 10))+1
    elif type in ['o','O',]:
        return int(math.log(num, 8))+1
    elif type in ['x','X',]:
        return int(math.log(num, 16))+1
    elif type in ['i','I',]:
        m = len(roman_max_widths)
        if num > m:
            update_roman_max_widths(num)
        return roman_max_widths[num]
    else:
        raise ValueError("Unexpected type argument to num_width(): '%s'" % (type, ))


ol_style_formatter = {
    None:   lambda n: tuple(),
    'a':    lambda n: tuple([int2letter(n)]),
    'A':    lambda n: tuple([int2letter(n).upper()]),
    'c':    lambda n: tuple([int2letter(n)]),
    'C':    lambda n: tuple([int2letter(n).upper()]),
    '1':    lambda n: tuple([n]),
    'd':    lambda n: tuple([n]),
    'i':    lambda n: tuple([int2roman(n)]),
    'I':    lambda n: tuple([int2roman(n).upper()]),
    'o':    lambda n: tuple(['%o'%n]),
    'O':    lambda n: tuple([('%o'%n).upper()]),
    'x':    lambda n: tuple(['%x'%n]),
    'X':    lambda n: tuple(['%X'%n]),
}




# Copyright The IETF Trust 2018, All Rights Reserved
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function, division

# Internal utility functions.  Not meant for public usage.

import base64
import re
import sys
import textwrap

from collections import OrderedDict
from lxml.etree import _Comment, _ProcessingInstruction

from urllib.request import quote

try:
    from xml2rfc import debug
    assert debug
except ImportError:
    pass

import xml2rfc.log


# ----------------------------------------------------------------------
# Text wrapping

class TextWrapper(textwrap.TextWrapper):
    """ Subclass that overrides a few things in the standard implementation """
    def __init__(self, **kwargs):
        textwrap.TextWrapper.__init__(self, **kwargs)

        # Override wrapping regex, preserve '/' before linebreak
        self.wordsep_re = re.compile(
            u'('
            u'[ \t\n\r\f\v]+|'                                  # any ASCII whitespace
            u'[^\\s-]*\\w+/(?=[A-Za-z]\\w*)|'                   # forward-slash separated words
            u'[^\\s-]*\\w+[^0-9\\s]-(?=\\w+[^0-9\\s])|'         # hyphenated words
            u'\u200b|'                                          # &zwsp; zero-width space is breakable space
            u'''(?<=[\\w\\!"'\\&\\.\\,\\?])-{2,}(?=\\w))'''     # em-dash
            u'(?![\u00A0|\u2060|\uE060])')                      # UNLESS &nbsp; or &wj; 

        self.wordsep_re_uni = re.compile(self.wordsep_re.pattern, re.U)

        # Override end of line regex, double space after '].'
        self.sentence_end_re = re.compile(r'[A-Za-z0-9\>\]\)\"\']'  # letter, end parentheses
                            r'[\.\!\?]'     # sentence-ending punct.
                            r'[\"\'\)\]]?'  # optional end-of-quote, end parentheses
                            r'\Z'           # end of chunk
                            )

        # Exceptions which should not be treated like end-of-sentence
        self.not_sentence_end_re = re.compile(
            # start of string or non-alpha character
            r'(^|\s)'
            r'('
            # Single uppercase letter, dot, enclosing parentheses or quotes
            r'\.[\]\)\'"]*'
            # Tla with leading uppercase, and special cases
            # (Note: v1 spelled out Fig, Tbl, Mrs, Drs, Rep, Sen, Gov, Rev, Gen, Col, Maj and Cap,
            #  but those are redundant with the Tla regex.)
            r'|([\(\[])?([A-Z][a-z][a-z]|Eq|[Cc]f|vs|resp|viz|ibid|[JS]r|M[rs]|Messrs|Mmes|Dr|Profs?|St|Lt|a\.k\.a|i\.e|e\.g)\.'
            r')\Z' # trailing dot, end of group and end of chunk
            )

        # Start of next sentence regex
        self.sentence_start_re = re.compile("^[\"'([]*[A-Z]")

    def _fix_sentence_endings(self, chunks):
        """_fix_sentence_endings(chunks : [string])

        Correct for sentence endings buried in 'chunks'.  Eg. when the
        original text contains "... foo.\nBar ...", munge_whitespace()
        and split() will convert that to [..., "foo.", " ", "Bar", ...]
        which has one too few spaces; this method simply changes the one
        space to two.
        """
        i = 0
        patsearch = self.sentence_end_re.search
        skipsearch = self.not_sentence_end_re.search
        startsearch = self.sentence_start_re.search
        while i < len(chunks)-2:
            if (chunks[i+1] == " " and patsearch(chunks[i])
                                  and skipsearch(chunks[i])==None
                                  and startsearch(chunks[i+2])):
                chunks[i+1] = "  "
                i += 2
            else:
                i += 1

    def wrap(self, text, initial='', subsequent_indent=None, width=None,
        fix_doublespace=True, fix_sentence_endings=True, drop_whitespace=True):
        """ Mirrored implementation of wrap which replaces characters properly
            and also lets you easily specify indentation on the fly
        """
        # Handle U+2028 LINE SEPARATOR by recursing
        if '\u2028' in text:
            parts = text.split('\u2028')
            wrapped = []
            for part in parts:
                part = part.lstrip()
                if part:
                    wrapped += self.wrap(part, initial=initial, subsequent_indent=subsequent_indent,
                                    width=width, fix_doublespace=fix_doublespace,
                                    fix_sentence_endings=fix_sentence_endings,
                                    drop_whitespace=drop_whitespace) 
                else:
                    wrapped.append(part)
                initial = ''
            return wrapped

        # Set indentation
        _width = self.width
        if width != None:
            self.width = width
        self.initial_indent = initial
        if subsequent_indent == None:
            self.subsequent_indent = initial
        else:
            self.subsequent_indent = subsequent_indent
        self.fix_sentence_endings = fix_sentence_endings
        self.drop_whitespace = drop_whitespace

        # --- Original implementation ----------------------------------
        text = self._munge_whitespace(text)
        # --- End original implementation ------------------------------

        # Maybe remove double (and more) spaces, except when they might be between sentences
        if fix_doublespace:
            text = re.sub("([^.!?])  +", r"\1 ", text)
            text = re.sub("([.!?])   +", r"\1  ", text)

        # prevent breaking "Section N.N" and "Appendix X.X"
        text = re.sub("(Section|Appendix|Figure|Table) ", u"\\1\u00A0", text)

        # Replace some characters after splitting has occurred
        # --- Original implementation ----------------------------------
        chunks = self._split(text)
        # --- End original implementation ------------------------------

        parts = chunks
        chunks = []
        max_word_len = self.width - len(self.subsequent_indent)
        for chunk in parts:
            # remove zero-width characters before looking at length
            chunk2 = re.sub('[\u200B\u200C\u2060\uE060]', '', chunk)
            if len(chunk2) > max_word_len:
                # No use trying not to break when we have a chunk longer
                # than available horizontal space.  Remove wj and nbsp and
                # try again.
                chunk2 = re.sub('[\u2060\uE060]', '', chunk).replace('\u00A0', ' ')
                bits = self._split(chunk2)
                for bit in bits:
                    chunk3 = re.sub('[\u2060\uE060]', '', bit).replace('\u00A0', ' ')
                    if len(chunk3) > max_word_len:
                        chunks += self._split(chunk3)
                    else:
                        chunks += [ chunk3 ]
            else:
                chunks += [ chunk2 ]
        
        # --- Original implementation ----------------------------------
        if self.fix_sentence_endings:
            self._fix_sentence_endings(chunks)
        wrapped = self._wrap_chunks(chunks)
        # --- End original implementation ------------------------------

        self.width = _width
        for i, chunk in enumerate(wrapped):
            wrapped[i] = chunk.replace(u'\uE060', '')
        return wrapped

    def fill(self, *args, **kwargs):
        kwargs.pop('fill', None)
        return "\u2028".join(self.wrap(*args, **kwargs))


class TextSplitter(textwrap.TextWrapper):
    """ Subclass that overrides a few things in the standard implementation """
    def __init__(self, **kwargs):
        hyphen_split = kwargs.pop('hyphen_split', False)
        textwrap.TextWrapper.__init__(self, **kwargs)

        # Override wrapping regex, preserve '/' before linebreak
        if hyphen_split:
            self.wordsep_re = re.compile(
                u'('
                u'[ \t\n\r\f\v]+|'                                  # any ASCII whitespace
                u'[^\\s-]*\\w+/(?=[A-Za-z]\\w*)|'                   # forward-slash separated words
                u'\u200b|'                                          # &zwsp; zero-width space is breakable space
                u'\u2028|'                                          # &br; unicode 'line separator'
                u'[^\\s-]*\\w+[^0-9\\s]-(?=\\w+)|'                  # hyphenated words
                u'''(?<=[\\w\\!"'\\&\\.\\,\\?])-{2,}(?=\\w))'''     # em-dash
                u'(?![\u00A0|\u2060|\uE060])')                      # UNLESS &nbsp; or &wj; 
        else:
            self.wordsep_re = re.compile(
                u'('
                u'[ \t\n\r\f\v]+|'                                  # any ASCII whitespace
                u'[^\\s-]*\\w+/(?=[A-Za-z]\\w*)|'                   # forward-slash separated words
                u'\u200b|'                                          # &zwsp; zero-width space is breakable space
                u'\u2028|'                                          # &br; unicode 'line separator'
                u'''(?<=[\\w\\!"'\\&\\.\\,\\?])-{2,}(?=\\w))'''     # em-dash
                u'(?![\u00A0|\u2060|\uE060])')                      # UNLESS &nbsp; or &wj; 

        self.wordsep_re_uni = re.compile(self.wordsep_re.pattern, re.U)


def justify_inline(left_str, center_str, right_str, width=72):
    """ Takes three string arguments and outputs a single string with the
        arguments left-justified, centered, and right-justified respectively.

        Raises a warning if the combined length of the three strings is
        greater than the width, and trims the longest string
    """
    strings = [left_str.rstrip(), center_str.strip(), right_str.strip()]
    sumwidth = sum( [len(s) for s in strings] )
    if sumwidth > width:
        # Trim longest string
        longest_index = strings.index(max(strings, key=len))
        xml2rfc.log.warn('The inline string was truncated because it was ' \
                         'too long:\n  ' + strings[longest_index])
        strings[longest_index] = strings[longest_index][:-(sumwidth - width)]

    if len(strings[1]) % 2 == 0:
        center = strings[1].center(width)
    else:
        center = strings[1].center(width+1)        
    right = strings[2].rjust(width)
    output = list(strings[0].ljust(width))
    for i, char in enumerate(output):
        if center[i] != ' ':
            output[i] = center[i]
        elif right[i] != ' ':
            output[i] = right[i]
    return ''.join(output)

# ----------------------------------------------------------------------
# Other text operations


def ascii_split(text):
    """ We have unicode strings, but we want to split only on the ASCII
        whitespace characters so that nbsp does not get split.
    """

    if isinstance(text, type('')):
        return text.split()
    return re.split("[ \t\n\r\f\v]+", text)


def urlkeep(text, max=72):
    """ Insert word join XML entities on forward slashes and hyphens
        in a URL so that it stays on one line
    """
    wj_char = u'\uE060'
    def replacer(match):
        url = match.group(0)
        if len(url) > max:
            return url
        else:
            return ( url.replace('/', '/' + wj_char) 
                        .replace('-', '-' + wj_char) 
                        .replace(':', ':' + wj_char)
                    )
    if 'http://' in text:
        text = re.sub(r'http:\S*', replacer, text)
    if 'https://' in text:
        text = re.sub(r'https:\S*', replacer, text)
    return text

# ----------------------------------------------------------------------
# Tree operations

def formatXmlWhitespace(tree):
    """ Traverses an lxml.etree ElementTree and properly formats whitespace
    
        We replace newlines with single spaces, unless it ends with a
        period then we replace the newline with two spaces.
    """
    for element in tree.iter():
        # Preserve formatting on artwork
        if element.tag != 'artwork':
            if element.text is not None:
                element.text = re.sub(r'\s*\n\s*', ' ', \
                               re.sub(r'\.\s*\n\s*', '.  ', \
                               element.text.lstrip()))

            if element.tail is not None:
                element.tail = re.sub(r'\s*\n\s*', ' ', \
                               re.sub(r'\.\s*\n\s*', '.  ', \
                               element.tail))


def safeReplaceUnicode(tree):
    """ Traverses an lxml.etree ElementTree and replaces unicode characters 
        with the proper equivalents specified in rfc2629-xhtml.ent.

        Writers should call this method if the entire RFC document needs to
        be ascii formatted
    """
    for element in tree.iter():
        if element.text:
            try:
                element.text = element.text.encode('ascii')
            except UnicodeEncodeError:
                element.text = _replace_unicode_characters(element.text)
        if element.tail:
            try:
                element.tail = element.tail.encode('ascii')
            except UnicodeEncodeError:
                element.tail = _replace_unicode_characters(element.tail)
        for key in element.attrib.keys():
            try:
                element.attrib[key] = element.attrib[key].encode('ascii')
            except UnicodeEncodeError:
                element.attrib[key] = \
                _replace_unicode_characters(element.attrib[key])


def safeTagSlashedWords(tree):
    """ Traverses an lxml.etree ElementTree and replace words separated
        by slashes if they are on the list
    """
    slashList = {}
    for element in _slash_replacements:
        slashList[element] = re.sub(u'/', u'\uE060/\uE060', element)
        
    for element in tree.iter():
        if element.tag == 'artwork':
            continue
        if element.text:
            element.text = _replace_slashed_words(element.text, slashList)
        if element.tail:
            element.tail = _replace_slashed_words(element.tail, slashList)
        # I do not know that this makes any sense
        # for key in element.attrib.keys():
        #    element.attrib[key] = \
        #        _replace_slashed_words(element.attrib[key], slashList)


def find_duplicate_ids(schema, tree):
    # get attributes specified with data type "ID"
    id_data = schema.xpath("/x:grammar/x:define/x:element//x:attribute/x:data[@type='ID']", namespaces=namespaces)
    attr = set([ i.getparent().get('name') for i in id_data ])
    # Check them one by one
    return find_duplicate_attr_values(attr, tree)

def find_duplicate_html_ids(tree):
    return find_duplicate_attr_values(['id',], tree)

def find_duplicate_attr_values(attr, tree):
    dups = []
    # Check them one by one
    for a in attr:
        seen = set()
        for e in tree.xpath('.//*[@%s]' % a):
            id = e.get(a)
            if id != None and id in seen:
                dups.append((a, id, e))
            else:
                seen.add(id)
    return dups

def strip_link_attachments(tree):
    """
    Find link tags with rel="attachment".
    """
    for attachment in tree.xpath('//link[@rel="attachment"]'):
        xml2rfc.log.warn(f"Removed {attachment}. link relationships type attachment is not allowed.")
        attachment.getparent().remove(attachment)


# ----------------------------------------------------------------------
# Unicode operations


def _replace_unicode_characters(str):
    """ replace those Unicode characters that we do not use internally
        &wj; &zwsp; &nbsp; &nbhy;
    """
    str = str.replace('\uE060', '\u2060')
    while True:
        match = re.search(u'([^ -\x7e\u2060\u200B\u00A0\u2011\r\n])', str)
        if not match:
            return str
        if match.group(1) in _unicode_replacements:
            str = re.sub(match.group(1), _unicode_replacements[match.group(1)], str)
        else:
            entity = match.group(1).encode('ascii', 'xmlcharrefreplace').decode('ascii')
            str = re.sub(match.group(1), entity, str)
            xml2rfc.log.warn('Illegal character replaced in string: ' + entity)


# Ascii representations of unicode chars from rfc2629-xhtml.ent
# Auto-generated from comments in rfc2629-xhtml.ent
_unicode_replacements = {
    u'\x09': ' ',
    u'\xa0': ' ',
    u'\xa1': '!',
    u'\xa2': '[cents]',
    u'\xa3': 'GBP',
    u'\xa4': '[currency units]',
    u'\xa5': 'JPY',
    u'\xa6': '|',
    u'\xa7': 'S.',
    u'\xa9': '(C)',
    u'\xaa': 'a',
    u'\xab': '<<',
    u'\xac': '[not]',
    u'\xae': '(R)',
    u'\xaf': '_',
    u'\xb0': 'o',
    u'\xb1': '+/-',
    u'\xb2': '^2',
    u'\xb3': '^3',
    u'\xb4': "'",
    u'\xb5': '[micro]',
    u'\xb6': 'P.',
    u'\xb7': '.',
    u'\xb8': ',',
    u'\xb9': '^1',
    u'\xba': 'o',
    u'\xbb': '>>',
    u'\xbc': '1/4',
    u'\xbd': '1/2',
    u'\xbe': '3/4',
    u'\xbf': '?',
    u'\xc0': 'A',
    u'\xc1': 'A',
    u'\xc2': 'A',
    u'\xc3': 'A',
    u'\xc4': 'Ae',
    u'\xc5': 'Ae',
    u'\xc6': 'AE',
    u'\xc7': 'C',
    u'\xc8': 'E',
    u'\xc9': 'E',
    u'\xca': 'E',
    u'\xcb': 'E',
    u'\xcc': 'I',
    u'\xcd': 'I',
    u'\xce': 'I',
    u'\xcf': 'I',
    u'\xd0': '[ETH]',
    u'\xd1': 'N',
    u'\xd2': 'O',
    u'\xd3': 'O',
    u'\xd4': 'O',
    u'\xd5': 'O',
    u'\xd6': 'Oe',
    u'\xd7': 'x',
    u'\xd8': 'Oe',
    u'\xd9': 'U',
    u'\xda': 'U',
    u'\xdb': 'U',
    u'\xdc': 'Ue',
    u'\xdd': 'Y',
    u'\xde': '[THORN]',
    u'\xdf': 'ss',
    u'\xe0': 'a',
    u'\xe1': 'a',
    u'\xe2': 'a',
    u'\xe3': 'a',
    u'\xe4': 'ae',
    u'\xe5': 'ae',
    u'\xe6': 'ae',
    u'\xe7': 'c',
    u'\xe8': 'e',
    u'\xe9': 'e',
    u'\xea': 'e',
    u'\xeb': 'e',
    u'\xec': 'i',
    u'\xed': 'i',
    u'\xee': 'i',
    u'\xef': 'i',
    u'\xf0': '[eth]',
    u'\xf1': 'n',
    u'\xf2': 'o',
    u'\xf3': 'o',
    u'\xf4': 'o',
    u'\xf5': 'o',
    u'\xf6': 'oe',
    u'\xf7': '/',
    u'\xf8': 'oe',
    u'\xf9': 'u',
    u'\xfa': 'u',
    u'\xfb': 'u',
    u'\xfc': 'ue',
    u'\xfd': 'y',
    u'\xfe': '[thorn]',
    u'\xff': 'y',
    u'\u0152': 'OE',
    u'\u0153': 'oe',
    u'\u0161': 's',
    u'\u0178': 'Y',
    u'\u0192': 'f',
    u'\u02dc': '~',
    u'\u2002': ' ',
    u'\u2003': ' ',
    u'\u2009': ' ',
    u'\u2013': '-',
    u'\u2014': u'-\u002D',
    u'\u2018': "'",
    u'\u2019': "'",
    u'\u201a': "'",
    u'\u201c': '"',
    u'\u201d': '"',
    u'\u201e': '"',
    u'\u2020': '*!*',
    u'\u2021': '*!!*',
    u'\u2022': 'o',
    u'\u2026': '...',
    u'\u2030': '[/1000]',
    u'\u2032': "'",
    u'\u2039': '<',
    u'\u203a': '>',
    u'\u2044': '/',
    u'\u20ac': 'EUR',
    u'\u2122': '[TM]',
    u'\u2190': '<-\u002D',
    u'\u2192': '\u002D->',
    u'\u2194': '<->',
    u'\u21d0': '<==',
    u'\u21d2': '==>',
    u'\u21d4': '<=>',
    u'\u2212': '-',
    u'\u2217': '*',
    u'\u2264': '<=',
    u'\u2265': '>=',
    u'\u2329': '<',
    u'\u232a': '>',

    # rfc2629-other.ent
    u'\u0021': '!',
    u'\u0023': '#',
    u'\u0024': '$',
    u'\u0025': '%',
    u'\u0028': '(',
    u'\u0029': ')',
    u'\u002a': '*',
    u'\u002b': '+',
    u'\u002c': ',',
    u'\u002d': '-',
    u'\u002e': '.',
    u'\u002f': '/',
    u'\u003a': ':',
    u'\u003b': ';',
    u'\u003d': '=',
    u'\u003f': '?',
    u'\u0040': '@',
    u'\u005b': '[',
    u'\u005d': ']',
    u'\u005e': '^',
    u'\u005f': '_',
    u'\u0060': '`',
    u'\u007b': '{',
    u'\u007c': '|',
    u'\u007d': '}',
    u'\u017d': 'Z',
    u'\u017e': 'z',
    u'\u2010': '-',
}

def parse_pi(pi, pis):
    """ Add a processing instruction to the current state 

        Will also return the dictionary containing the added instructions
        for use in things like ?include instructions
    """
    if pi.text:
        # Split text in the format 'key="val"'
        chunks = re.split(r'=[\'"]([^\'"]*)[\'"]', pi.text)
        # Create pairs from this flat list, discard last element if odd
        tmp_dict = dict(zip(chunks[::2], chunks[1::2]))
        for key, val in tmp_dict.items():
            # Update main PI state
            pis[key] = val
        # Return the new values added
        return tmp_dict
    return {}


def _replace_slashed_words(str, slashList):
    """ replace slashed separated words the list with
        <word1> &nbsp; / &nbsp; <word 2>
    """
    match = re.findall(r'(\w+(/\w+)+)', str)
    for item in match:
        if item[0] in slashList:
            str = re.sub(item[0], slashList[item[0]], str)
    return str


_slash_replacements = [
    u'PDF/A',
    u'PDF/UA',
    u'PDF/VT',
    u'S/MIME',
    # If you remove me the regression test fails
    u'this/is/a/long/test',
];

            
def build_dataurl(mime, data, base64enc=False):
    if base64enc:
        data = quote(base64.b64encode(data), safe='/~')
        url = "data:%s;base64,%s" % (mime, data)
    else:
        data = quote(data, safe='/~')
        url = "data:%s,%s" % (mime, data)
    return url

namespaces={
    'x':    'http://relaxng.org/ns/structure/1.0',
    'a':    'http://relaxng.org/ns/compatibility/annotations/1.0',
    'xi':   'http://www.w3.org/2001/XInclude',
    'svg':  'http://www.w3.org/2000/svg',
    'xml':  'http://www.w3.org/XML/1998/namespace',
    'xlink':'http://www.w3.org/1999/xlink',
}


def sdict(d):
    "Create an ordered dict from the given dict, with sorted initial insertion"
    # For python 3.6 and later, lxml obeys dictionary insertion order for
    # attributes, for earlier it sorts initial attributes, so we need to
    # use a regular dict for 3.6 and later
    if sys.version_info.major == 3 and sys.version_info.minor >= 6:
        return dict( (k, d[k]) for k in sorted(list(d.keys())) )
    else:
        if isinstance(d, OrderedDict):
            return d
        return OrderedDict( (k, d[k]) for k in sorted(list(d.keys())) )

# ----------------------------------------------------------------------
# Element operations


def isempty(e):
    "Return True if e has no children and no text content"
    return not (len(e) or (e.text and e.text.strip()) or (e.tail and e.tail.strip()))

def isblock(e):
    "Return True if e is a block level element"
    from xml2rfc.writers.base import block_tags
    return e.tag in block_tags

def iscomment(e):
    "Return True if e is a comment or PI"
    return isinstance(e, (_Comment, _ProcessingInstruction))

def hastext(e, ignore=[]):
    "Return a list of text-level immediate children"
    head = [ e.text ] if e.text and e.text.strip() else []
    ignore = set(ignore + ['t'])
    items = ( head + [ c for c in e.iterchildren() if not (isblock(c) or iscomment(c) or (c.tag in ignore))]
                   + [ c.tail for c in e.iterchildren() if c.tail and c.tail.strip() ] )
    return items

def clean_text(s):
    """Replace internal use code points and various other special code points with plain equivalents"""
    # spaces
    s = re.sub(r'[\u00a0\u2028]', ' ', s)
    # hyphens
    s = s.replace(r'\u2011', '-')
    # zero-width
    s = re.sub(r'[\u200B\u2060\ue060]', '', s)
    return s.strip()

def is_htmlblock(h):
    return h.tag in set([ 'address', 'article', 'aside', 'blockquote', 'div', 'dl', 'figure',
        'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'nav', 'ol', 'p', 'pre', 'script', 'section',
        'table', 'ul', ])

def slugify(s):
    s = s.strip().lower()
    s = re.sub(r'[^\w\s/|@=-]', '', s)
    s = re.sub(r'[-_\s/|@=]+', '_', s)
    s = s.strip('_')
    return s



# --------------------------------------------------
# Copyright The IETF Trust 2011, All Rights Reserved
# --------------------------------------------------

# External libs
from lxml.html import builder as E
import lxml
import os.path
import string
import sys
import datetime
import calendar

from html import escape

try:
    from xml2rfc import debug
    debug.debug = True
    assert debug
except ImportError:
    pass

# Local libs
import xml2rfc
from xml2rfc.util.name import short_author_name_parts
from xml2rfc.util.unicode import expand_unicode_element
from xml2rfc.writers.base import BaseRfcWriter, default_options



class HtmlRfcWriter(BaseRfcWriter):
    """ Writes to an HTML file.

        Uses HTML templates located in the templates directory.
    """
    # HTML Specific Defaults that are not provided in templates or XML
    html_defaults = { 
        'references_url': 'https://tools.ietf.org/html/',
    }

    def __init__(self, xmlrfc, quiet=None, options=default_options,
                               date=None, templates_dir=None):
        if not quiet is None:
            options.quiet = quiet
        BaseRfcWriter.__init__(self, xmlrfc, options=options)
        self.list_counters = {}
        self.iref_index = []
        self.nbws_cond = ' '    # Use space as they are in an anchor element anyway
        self.cref_counter = 0
        
        # Initialize templates directory
        self.template_dir = templates_dir or \
                             os.path.join(os.path.dirname(xml2rfc.__file__),
                                          'templates')

        # Open template files and read into string.template objects
        self.templates = {}
        for filename in ['base.html',
                         'address_card.html']:
            with open(os.path.join(self.template_dir, filename), 'r') as file:
                self.templates[filename] = string.Template(file.read())
        # Buffers to aggregate various elements before processing template
        self.buffers = {'front': [],
                        'body': [],
                        'header_rows': [],
                        'toc_head_links': [],
                        'toc_rows': []}
        
        # Use an active buffer pointer
        self.buf = self.buffers['front']
        
        # String holding the final document to write
        self.output = ''
        
        # Temporary div for recursive functions
        self.temp_div = E.DIV()

        # Table to insert the iref into
        self.iref_table = None

    def _add_iref_to_index(self, element):
        # Add anchor to index
        item = element.attrib.get('item', None)
        if item:
            subitem = element.attrib.get('subitem', None)
            anchor = '.'.join(('index', item))
            if subitem:
                anchor += '.' + subitem
            # Create internal iref
            self._make_iref(item, subitem=subitem, anchor=anchor)
            # Create anchor element
            a = E.A(name=anchor)
            if element.tail:
                a.tail = element.tail
            return [a]

    def write_list(self, list, parent, level=0):
        # grab the anchors that exist
        if self.indexmode:
            t_count = 0
            for t in list.findall('t'):
                t_count += 1
                anchor = t.attrib.get('anchor')
                if anchor:
                    self._indexListParagraph(t_count, anchor)
        # style comes from the node if one exists
        style = list.attrib.get('style', '')
        if not style:
            # otherwise look for the nearest list parent with a style and use it
            for elder in list.iterancestors():
                if elder.tag == 'list':
                    style = elder.attrib.get('style', '')
                    if style:
                        break
        if not style:
            style = 'empty'
        if style == 'hanging':
            list_elem = E.DL()
            hangIndent = list.attrib.get('hangIndent', '8')
            style = 'margin-left: ' + hangIndent
            for t in list.findall('t'):
                hangText = t.attrib.get('hangText', '')
                dt = E.DT(hangText)
                dd = E.DD(style=style)
                list_elem.append(dt)
                list_elem.append(dd)
                self.write_t_rec(t, parent=dd, level=level +1)
        elif style.startswith('format'):
            format_str = style.partition('format ')[2]
            allowed_formats = ('%c', '%C', '%d', '%i', '%I', '%o', '%x', '%X')
            if not any([ f in format_str for f in allowed_formats]):
                xml2rfc.log.warn('Invalid format specified: %s ' 
                                 '(Must be one of %s)' % (style,
                                    ', '.join(allowed_formats)))
            counter_index = list.attrib.get('counter', None)
            if not counter_index:
                counter_index = 'temp' + str(level)
                self.list_counters[counter_index] = 0
            elif counter_index not in self.list_counters:
                # Initialize if we need to
                self.list_counters[counter_index] = 0
            list_elem = E.DL()
            for t in list.findall('t'):
                self.list_counters[counter_index] += 1
                count = self.list_counters[counter_index]
                bullet = self._format_counter(format_str, count) + ' '
                dt = E.DT(bullet)
                dd = E.DD()
                list_elem.append(dt)
                list_elem.append(dd)
                self.write_t_rec(t, parent=dd, level=level + 1)
        else:
            if style == 'symbols':
                list_elem = E.UL()
            elif style == 'numbers':
                list_elem = E.OL()
            elif style == 'letters':
                letter_style = "list-style-type: upper-alpha" if (level % 2) else "list-style-type: lower-alpha"
                list_elem = E.OL(style= letter_style)
            else:  # style == empty
                list_elem = E.UL()
                list_elem.attrib['class'] = 'empty'
            for t in list.findall('t'):
                li = E.LI()
                list_elem.append(li)
                self.write_t_rec(t, parent=li, level= level + 1)
        parent.append(list_elem)

    def _create_toc(self):
        # Add link for toc itself
        link = E.LINK(href='#rfc.toc')
        link.attrib["rel"] = 'Contents'
        self.buffers['toc_head_links'].append(self._serialize(link))
        tocdepth = self.pis['tocdepth']
        try:
            tocdepth = int(tocdepth)
        except ValueError:
            xml2rfc.log.warn('Invalid tocdepth specified, must be integer:', \
                             tocdepth)
            tocdepth = 3
        curdepth = 1
        for item in self._getTocIndex():
            if item.level <= tocdepth:
                # set UL level
                if item.level > curdepth:
                    self.buffers['toc_rows'].append('<ul>')
                for i in range(item.level, curdepth):
                    self.buffers['toc_rows'].append('</ul>')
                curdepth = item.level

                # Create link for head
                link = E.LINK(href='#' + item.autoAnchor)
                link.attrib['rel'] = 'copyright' in item.autoAnchor and \
                                     'Copyright' or 'Chapter'
                if item.title and item.counter:
                    link.attrib['title'] = item.counter + ' ' + item.title.strip()
                self.buffers['toc_head_links'].append(self._serialize(link))
                # Create actual toc list item
                a = E.A(item.title.strip(), href='#' + item.autoAnchor)
                counter_text = item.counter and item.counter + '.   ' or ''
                # Prepend appendix at first level
                if item.level == 1 and item.appendix:
                    counter_text = "Appendix " + counter_text
                li = E.LI(counter_text)
                li.append(a)
                self.buffers['toc_rows'].append(self._serialize(li))
        for i in range(1, curdepth):
            self.buffers['toc_rows'].append('</ul>')

    def _serialize(self, element):
        if sys.version > '3':
            return lxml.html.tostring(element, pretty_print=True, method='html', encoding='utf-8').decode('utf-8')
        else:
            return lxml.html.tostring(element, pretty_print=True, method='html')
    
    def _flush_temp_div(self):
        lines = []
        for child in self.temp_div:
            lines.append(self._serialize(child))
            child.drop_tree()
        return '\n'.join(lines)
    
    def _expand_ref(self, element):
        """ Return a list of HTML elements that represent the reference """   
        if element.tag == 'xref':
            target = element.attrib.get('target', '')
            format = element.attrib.get('format', self.defaults['xref_format'])
            item = self._getItemByAnchor(target)
            if not self.indexmode:
                if not item:
                    xml2rfc.log.warn("Can't resolve xref target %s" % target)
                else:
                    item.used = True
            # Create xref from index lookup
            if not item:
                text = '[' + target + ']'
            elif format == 'none':
                text = ''
            elif format == 'counter':
                text = item.counter
            elif format == 'title':
                text = item.title.strip() if item.title else ''
            else:
                # Default
                text = item.autoName

            # following the V3 HTML -
            #  If you specify text, that is what you get.
            if element.text:
                text = element.text.rstrip()
             
            a = E.A(href='#' + target)
            a.attrib["class"] = "xref"
            a.text = text
            if element.tail:
                a.tail = element.tail
            
            return [a]

        elif element.tag == 'eref':
            target = element.attrib.get('target', '')
            if element.text:
                a = E.A(element.text, href=target)
                a.tail = element.tail
                return [a]
            else:
                sp1 = E.SPAN('<')
                a = E.A(target, href=target)
                sp2 = E.SPAN('>')
                sp2.tail = element.tail
                return [sp1, a, sp2]
        elif element.tag == 'cref':
            self.cref_counter += 1
            anchor = element.attrib.get('anchor', None)
            if anchor is None:
                anchor = 'CREF' + str(self.cref_counter)
            a = E.A('[' + anchor + ']', id=anchor)
            a.attrib['class'] = 'info'
            source = element.attrib.get('source', '')
            if source:
                source = source + ": "
            b = E.SPAN(source + element.text)
            b.attrib['class'] = 'info'
            a.append( b )
            self._indexCref(self.cref_counter, anchor)
            if element.tail:
                a.tail = element.tail
            return [a]
        elif element.tag == 'iref':
            return self._add_iref_to_index(element)
        elif element.tag == 'spanx':
            style = element.attrib.get('style', self.defaults['spanx_style'])
            text = ''
            if element.text:
                text = element.text
            elem = None
            if style == 'strong':
                elem = E.STRONG(text)
            elif style == 'verb':
                elem = E.SAMP(text)
            else:
                # Default to style=emph
                elem = E.EM(text)
            if element.tail:
                elem.tail = element.tail
            return [elem]

    def _expand_u(self, element):
        try:
            text = expand_unicode_element(element)
        except (RuntimeError, ValueError) as exc:
            text = ''
            xml2rfc.log.error('%s'%exc)
        if text:
            span = E.SPAN(text)
            span.set('class', 'unicode')
            return [span]
        return []

    # -----------------------------------------
    # Base writer overrides
    # -----------------------------------------

    def insert_toc(self):
        # We don't actually insert the toc here, but we swap the active buffer
        # So that the template replacement is correct
        self.buf = self.buffers['body']

    def write_raw(self, text, align='left', blanklines=0, delimiter=None, source_line=None):
        if text:
            # Add padding with delimiter/blanklines, if specified
            edge = delimiter and [delimiter] or ['']
            edge.extend([''] * blanklines)
            fill = '\n'.join(edge)
            fill += text
            edge.reverse()
            fill += '\n'.join(edge)

            # Run through template, add to body buffer
            pre = E.PRE(fill)
            self.buf.append(self._serialize(pre))

    def write_label(self, text, type='figure', align='center', source_line=None):
        # Ignore labels for table, they are handled in draw_table
        if type == 'figure':
            p = E.P(text)
            p.attrib['class'] = 'figure'
            # Add to body buffer
            self.buf.append(self._serialize(p))

    def write_heading(self, text, bullet=None, autoAnchor=None, anchor=None,
                      level=1):
        # Use a hierarchy of header tags if docmapping set
        h = E.H1()
        if self.pis['docmapping'] == 'yes':
            if level == 2:
                h = E.H2()
            elif level == 3:
                h = E.H3()
            elif level >= 4:
                h = E.H4()
        if autoAnchor:
            h.attrib['id'] = autoAnchor
        if bullet:
            # Use separate elements for bullet and text
            a_bullet = E.A(bullet)
            a_bullet.tail = ' '
            if autoAnchor:
                a_bullet.attrib['href'] = '#' + autoAnchor
            h.append(a_bullet)
            if anchor:
                # Use an anchor link for heading
                a_text = E.A(text.strip(), href='#' + anchor, )
                a_text.attrib["id"] = anchor
                h.append(a_text)
            else:
                # Plain text
                a_bullet.tail += text
        else:
            # Only use one <a> pointing to autoAnchor
            a = E.A(text)
            if autoAnchor:
                a.attrib['href'] = '#' + autoAnchor
            h.append(a)
        
        # Add to body buffer
        self.buf.append(self._serialize(h))

    def write_paragraph(self, text, align='left', autoAnchor=None):
        if text:
            p = E.P(text)
            if autoAnchor:
                p.attrib['id'] = autoAnchor        
            # Add to body buffer
            self.buf.append(self._serialize(p))

    def write_t_rec(self, t, autoAnchor=None, align='left', parent=None, level=0):
        """ Recursively writes a <t> element

            If no parent is specified, a dummy div will be treated as the parent
            and any text will go in a <p> element.  Otherwise text and
            child elements will be insert directly into the parent -- meaning
            we are in a list
        """
        if parent is None:
            parent = self.temp_div # Dummy div
            current = E.P()
            parent.append(current)
        else:
            current = parent
        if t.text:
            if "anchor" in t.attrib:
                a = E.A(name=t.attrib["anchor"])
                a.tail = t.text
                current.append(a)
            else:
                current.text = t.text
            if autoAnchor:
                current.attrib['id'] = autoAnchor
        for child in t:
            if child.tag in ['xref', 'eref', 'iref', 'cref', 'spanx']:
                for element in self._expand_ref(child):
                    current.append(element)
            elif child.tag == 'u':
                for element in self._expand_u(child):
                    current.append(element)
            elif child.tag == 'vspace':
                br = E.BR()
                current.append(br)
                blankLines = int(child.attrib.get('blankLines',
                                 self.defaults['vspace_blanklines']))
                for i in range(blankLines):
                    br = E.BR()
                    current.append(br)
                if child.tail:
                    br.tail = child.tail
            elif child.tag == 'list':
                self.write_list(child, parent, level=level)
                if child.tail:
                    parent.append(E.P(child.tail))
            elif child.tag == 'figure':
                # Callback to base writer method
                self.write_figure(child)
            elif child.tag == 'texttable':
                # Callback to base writer method
                self.write_table(child)
        # If we are back at top level, serialize the whole temporary structure
        # Add to body buffer
        if parent == self.temp_div:
            self.buf.append(self._flush_temp_div())

    def write_top(self, left_header, right_header):
        """ Buffers the header table """
        for i in range(max(len(left_header), len(right_header))):
            if i < len(left_header):
                left = left_header[i]
            else:
                left = ''
            if i < len(right_header):
                right = right_header[i]
            else:
                right = ''
                
            # Add row to header_rows buffer
            left_td = E.TD(left)
            left_td.attrib['class'] = 'left'
            right_td = E.TD(right)
            right_td.attrib['class'] = 'right'
            tr = E.TR(left_td, right_td)
            self.buffers['header_rows'].append(self._serialize(tr))

    def write_address_card(self, author):
        # Create substitution dictionary with empty string values
        vars = ['fullname', 'surname', 'role', 'organization',
                'street_spans', 'locality', 'region', 'code', 
                'locality_sep', 'region_sep', 'country', 'contact_spans']
        subs = dict(zip(vars, [''] * len(vars)))
        
        # Get values with safe defaults
        subs['fullname'] = author.attrib.get('fullname', '')
        subs['surname'] = author.attrib.get('surname', '')
        subs['role'] = author.attrib.get('role', '')
        subs['role'] = subs['role'] and "(%s)" % subs['role'] or ''
        subs['organization'] = author.find('organization') is not None and \
                               author.find('organization').text or ''
        address = author.find('address')
        
        # Crawl through optional elements
        if address is not None:
            postal = address.find('postal')
            if postal is not None:
                street_spans = []
                for street in postal.findall('street'):
                    if street.text:
                        span = E.SPAN(street.text)
                        span.attrib['class'] = 'vcardline'
                        street_spans.append(self._serialize(span))
                subs['street_spans'] = ''.join(street_spans)
                city = postal.find('city')
                if city is not None and city.text:
                    subs['locality'] = city.text
                    subs['locality_sep'] = ', '
                region = postal.find('region')
                if region is not None and region.text:
                    subs['region'] = region.text
                    subs['region_sep'] = ' '
                code = postal.find('code')
                if code is not None and code.text:
                    subs['code'] = code.text
                country = postal.find('country')
                if country is not None and country.text:
                    subs['country'] = country.text

            # Use temp div for contact spans
            contact_div = self.temp_div
            phone = address.find('phone')
            if phone is not None and phone.text:
                span = E.SPAN('Phone: ' + phone.text)
                span.attrib['class'] = 'vcardline'
                contact_div.append(span)
            facsimile = address.find('facsimile')
            if facsimile is not None and facsimile.text:
                span = E.SPAN('Fax: ' + facsimile.text)
                span.attrib['class'] = 'vcardline'
                contact_div.append(span)
            email = address.find('email')
            if email is not None and email.text:
                span = E.SPAN('EMail: ')
                span.attrib['class'] = 'vcardline'
                if self.pis['linkmailto'] == 'yes':
                    span.append(E.A(email.text, href='mailto:' + email.text))
                else:
                    span.text += email.text
                contact_div.append(span)
            uri = address.find('uri')
            if uri is not None and uri.text:
                span = E.SPAN('URI: ')
                span.attrib['class'] = 'vcardline'
                span.append(E.A(uri.text, href=uri.text))
                contact_div.append(span)

            # Serialize the temp div
            subs['contact_spans'] = self._flush_temp_div() 

        # Run through template and add to body buffer
        html = self.templates['address_card.html'].substitute(subs)
        self.buf.append(html)

    def write_reference_list(self, list):
        tbody = E.TBODY()
        refkeys = []
        refdict = {}
        for i, reference in enumerate(list.findall('reference')):
            tr = E.TR()
            # Use anchor or num depending on PI
            anchor = reference.attrib.get('anchor', str(i + self.ref_start))
            if self.pis['symrefs'] == 'yes':
                bullet = anchor
            else:
                bullet = str(i + self.ref_start)
            bullet_td = E.TD(E.B('[' + bullet + ']', id=anchor))
            bullet_td.attrib['class'] = 'reference'
            ref_td = E.TD()
            ref_td.attrib['class'] = 'top'
            last = None
            authors = reference.findall('front/author')
            for j, author in enumerate(authors):
                organization = author.find('organization')
                email = author.find('address/email')
                initials, surname = short_author_name_parts(author)
                initials = self.get_initials(author) or initials or ''
                a = None
                if j == len(authors) - 1 and len(authors) > 1:
                    last.tail = ' and '
                if surname is not None:
                    if j == len(authors) - 1 and len(authors) > 1:
                        # Last author, render in reverse
                        if len(initials) > 0:
                            name_string = initials + ' ' + surname
                        else:
                            name_string = surname
                    else:
                        if len(initials) > 0:
                            name_string = surname + ', ' + initials
                        else:
                            name_string = surname
                    a = E.A(name_string)
                    if email is not None and email.text:
                        if self.pis['linkmailto'] == 'yes':
                            a.attrib['href'] = 'mailto:' + email.text
                    if organization is not None and organization.text:
                        a.attrib['title'] = organization.text.strip()
                elif organization is not None and organization.text:
                    # Use organization instead of name
                    a = E.A(organization.text.strip())
                else:
                    continue
                ref_td.append(a)
                last = a
                a.tail = ', '
            title = reference.find('front/title')
            if title is not None and title.text:
                title_string = title.text.strip()
            else:
                xml2rfc.log.warn('No title specified in reference', \
                                 reference.attrib.get('anchor', ''))
                title_string = ''
            if title_string:
                if reference.attrib.get("quote-title", "true") == "true": # attribute default value: yes
                    if last is not None:
                        last.tail = ', "' 
                    else:
                        ref_td.text = '"'
                    title_a = E.A(title_string)
                    title_a.tail = '"'
                else:
                    if last is not None:
                        last.tail = ', '
                    title_a = E.A(title_string)
                    title_a.tail = ''
                ref_td.append(title_a)
            for seriesInfo in reference.findall('seriesInfo'):
                # Create title's link to document from seriesInfo
                if seriesInfo.attrib.get('name', '') == 'RFC':
                    title_a.attrib['href'] = \
                        self.html_defaults['references_url'] + \
                        'rfc' + seriesInfo.attrib.get('value', '')
                elif seriesInfo.attrib.get('name', '') == 'Internet-Draft':
                    title_a.attrib['href'] = \
                        self.html_defaults['references_url'] + \
                        seriesInfo.attrib.get('value', '')
                title_a.tail += ', '+seriesInfo.attrib.get('name', '') + ' ' + \
                             seriesInfo.attrib.get('value', '')
            if not title_a.attrib.has_key("href"):
                href = reference.attrib.get("target", None)
                if href:
                    title_a.attrib['href'] = href
            date = reference.find('front/date')
            if date is not None:
                month = date.attrib.get('month', '')
                if month.isdigit():
                    month = calendar.month_name[int(month)]
                year = date.attrib.get('year', '')
                if month or year:
                    title_a.tail += ', '
                    if month:
                        month += ' '
                    title_a.tail += month + year + '.'
            tr.append(bullet_td)
            tr.append(ref_td)
            refdict[bullet] = tr
            refkeys.append(bullet)
            # Render annotation as a separate paragraph
            annotation = reference.find('annotation')
            if annotation is not None and annotation.text:
                p = E.P()
                self.write_t_rec(annotation, parent=p)
                ref_td.append(p)
                    
        if self.pis['sortrefs'] == 'yes' and self.pis['symrefs'] == 'yes':
            refkeys.sort(key=str.lower)
        for key in refkeys:
            tbody.append(refdict[key])
        self.ref_start += i + 1                
        # Add to body buffer
        self.buf.append(self._serialize(E.TABLE(tbody)))

    def draw_table(self, table, table_num=None):
        style = 'tt %s %s' % ( table.attrib.get('style', self.defaults['table_style']),
                               table.attrib.get('align', self.defaults['table_align']) )
        cellpadding = '3'
        cellspacing = '0'
        htmltable = E.TABLE(cellpadding=cellpadding, )
        htmltable.attrib['cellspacing'] = cellspacing
        htmltable.attrib['class'] = style

        # Add caption, if it exists
        if 'title' in table.attrib and table.attrib['title']:
            caption = ''
            if table_num and self.pis['tablecount'] == 'yes':
                caption = 'Table ' + str(table_num) + ': '
            htmltable.append(E.CAPTION(caption + table.attrib['title']))

        # Draw headers
        header_row = E.TR()
        col_aligns = []
        for header in table.findall('ttcol'):
            th = E.TH()
            if header.text:
                th.text = header.text
            header_align = header.attrib.get('align',
                                             self.defaults['ttcol_align'])
            th.attrib['class'] = header_align
            # Store alignment information
            col_aligns.append(header_align)
            header_row.append(th)
        htmltable.append(E.THEAD(header_row))

        # Draw body
        body = E.TBODY()
        tr = E.TR()
        num_columns = len(col_aligns)
        for i, cell in enumerate(table.findall('c')):
            col_num = i % num_columns
            if col_num == 0 and i != 0:
                # New row
                body.append(tr)
                tr = E.TR()
            td = E.TD()
            if cell.text:
                # Add text
                td.text = cell.text
            for child in cell:
                # Add any inline elements (references)
                for element in self._expand_ref(child):
                    td.append(element)
            # Get alignment from header
            td.attrib['class'] = col_aligns[col_num]
            tr.append(td)
        body.append(tr)  # Add final row
        htmltable.append(body)

        # Add to body buffer
        self.buf.append(self._serialize(htmltable))

    def insert_anchor(self, text):
        # Add to body buffer
        self.buf.append(self._serialize(E.DIV(id=text)))

    def insert_iref_index(self):
        # Write the heading
        self.write_heading('Index', autoAnchor='rfc.index')
        table = E.TABLE()
        # Sort iref items alphabetically, store by first letter 
        alpha_bucket = {}
        keys = list(self._iref_index.keys())
        keys.sort(key=str.upper)
        for key in keys:
            letter = key[0].upper()
            if letter in alpha_bucket:
                alpha_bucket[letter].append(key)
            else:
                alpha_bucket[letter] = [key]
        for letter in sorted(alpha_bucket.keys()):
            # Add letter element
            table.append(E.TR(E.TD(E.STRONG(letter))))
            for item in sorted(alpha_bucket[letter]):
                # Add item element
                anchor = self._iref_index[item].anchor or ''
                if anchor:
                    anchor = '#' + anchor
                    td = E.TD(E.A(item, href=anchor))
                else:
                    td = E.TD(item)
                table.append(E.TR(E.TD(' '), td))
                subkeys = list(self._iref_index[item].subitems.keys())
                subkeys.sort(key=str.upper)
                for name in subkeys:
                    subitem = self._iref_index[item].subitems[name]
                    # Add subitem element
                    td = E.TD()
                    td.text = (u'\u00a0\u00a0')  # Spaces
                    anchor = subitem.anchor or ''
                    anchor = '#' + anchor
                    td.append(E.A(name, href=anchor))
                    table.append(E.TR(E.TD(' '), td))

        self.buf.append(self._serialize(table))
    
    def pre_indexing(self):
        pass

    def pre_rendering(self):
        # Reset buffers
        self.buffers = {'front': [],
                        'body': [],
                        'header_rows': [],
                        'toc_head_links': [],
                        'toc_rows': []}
        self.list_counters = {}
        self.buf = self.buffers['front']
        self.cref_counter = 0

    def post_rendering(self):
        # Create table of contents buffers
        self._create_toc()
        
        # Grab values that haven't been inserted yet
        background_image = (self.pis['background'] and 
            "background-image:url('%s');" % self.pis['background'] or '')
        title = self.r.find('front/title').text or ''
        
        if self.rfcnumber:
            docName = ''
        else:
            docName = self.r.attrib.get('docName', '') 
        description = ''
        abs = self.r.find('front/abstract')
        if abs != None:
            description = ''.join(abs.itertext())
        keywords = self.r.findall('front/keyword')
        keyword_list = [keyword.text for keyword in keywords if keyword.text]
        generator = "xml2rfc version %s - https://tools.ietf.org/tools/xml2rfc" % xml2rfc.__version__

        # Build ISO8601 date string
        docDate = ''
        date = self.r.find('front/date')
        if date is not None:
            # We can assume AT LEAST year was defined in base.py from _format_date()
            docDate = date.attrib.get('year', '')
            try:
                month = datetime.datetime.strptime(date.attrib.get('month', ''), '%B').month
                docDate += "-%02d" % int(month)
            except ValueError:
                pass
            if date.attrib.get('day'):
                docDate += "-%02d" % int(date.attrib.get('day'))

        # Build author string
        authors = self.r.findall('front/author')
        authors = [ a for a in authors if a.get('role') != 'contributor' ]
        authors = self._format_author_string(authors)

        # Run through base template, store in main output string
        subs = { 
                 # Replace flat values 
                 'background':      background_image,

                 # HTML-escaped values
                 'docName':         escape(docName, quote=True),
                 'docDate':         escape(docDate, quote=True),
                 'description':     escape(description, quote=True),
                 'generator':       escape(generator, quote=True),
                 'authors':         escape(authors, quote=True),
                 'keywords':        escape(', '.join(keyword_list), quote=True),
                 'title':           escape(title, quote=False),
                 
                 # Replace buffers
                 'front':           ''.join(self.buffers['front']),
                 'body':            ''.join(self.buffers['body']),
                 'header_rows':     ''.join(self.buffers['header_rows']),
                 'toc_head_links':  ''.join(self.buffers['toc_head_links']),
                 'toc_rows':        ''.join(self.buffers['toc_rows'])
                }
        for key in subs.keys():
            subs[key] = subs[key].replace('\uE060', '\u2060').encode('ascii', 'xmlcharrefreplace').decode('ascii')
        self.output = self.templates['base.html'].substitute(subs)

    def post_process_lines(self, lines):
        return lines

    def write_to_file(self, file):
        file.write(self.output)

#-------------------------------------------------------------------------------
# Unused methods
#-------------------------------------------------------------------------------

    def write_title(self, title, docName=None, source_line=None):
        pass



#         'trust200902',
#         'noModificationTrust200902',
#         'noDerivativesTrust200902',
#         'pre5378Trust200902',
#         'trust200811',
#         'noModificationTrust200811',
#         'noDerivativesTrust200811',





boilerplate_tlp = {
    "2.0": {
        # From https://trustee.ietf.org/license-info/IETF-TLP-2.htm
        "draft": [
            """<t>
            This Internet-Draft is submitted in full conformance
            with the provisions of BCP 78 and BCP 79.
            </t>""",
        ],
        "n/a": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as
            the document authors.  All rights reserved. 
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents in effect on the date of
            publication of this document (http://trustee.ietf.org/license-info).
            Please review these documents carefully, as they describe your rights
            and restrictions with respect to this document.
            </t>""",
        ],
        "noModification": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, except to format it for publication as an RFC or
            to translate it into languages other than English.
            </t>""",
        ],
        "noDerivatives": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, and it may not be published except as an
            Internet-Draft.
            </t>""",
        ],
        "pre5378": [
            """<t>
            This document may contain material from IETF Documents or IETF
            Contributions published or made publicly available before November
            10, 2008. The person(s) controlling the copyright in some of this
            material may not have granted the IETF Trust the right to allow
            modifications of such material outside the IETF Standards Process.
            Without obtaining an adequate license from the person(s)
            controlling the copyright in such materials, this document may not
            be modified outside the IETF Standards Process, and derivative
            works of it may not be created outside the IETF Standards Process,
            except to format it for publication as an RFC or to translate it
            into languages other than English.
            </t>""",
        ],
    },
    "3.0": {
        # From https://trustee.ietf.org/license-info/IETF-TLP-3.htm
        "draft": [
            """<t>
            This Internet-Draft is submitted to IETF in full conformance with
            the provisions of BCP 78 and BCP 79.
            </t>""",
        ],
        "n/a": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as
            the document authors.  All rights reserved. 
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents
            (http://trustee.ietf.org/license-info) in effect on the date of
            publication of this document. Please review these documents carefully,
            as they describe your rights and restrictions with respect to this
            document.  Code Components extracted from this document must include
            Simplified BSD License text as described in Section 4.e of the Trust
            Legal Provisions and are provided without warranty as described in the
            BSD License.
            </t>""",
        ],
        "noModification": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, except to format it for publication as an RFC or
            to translate it into languages other than English.
            </t>""",
        ],
        "noDerivatives": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, and it may not be published except as an
            Internet-Draft.
            </t>""",
        ],
        "pre5378": [
            """<t>
            This document may contain material from IETF Documents or IETF
            Contributions published or made publicly available before November
            10, 2008. The person(s) controlling the copyright in some of this
            material may not have granted the IETF Trust the right to allow
            modifications of such material outside the IETF Standards Process.
            Without obtaining an adequate license from the person(s)
            controlling the copyright in such materials, this document may not
            be modified outside the IETF Standards Process, and derivative
            works of it may not be created outside the IETF Standards Process,
            except to format it for publication as an RFC or to translate it
            into languages other than English.
            </t>""",
        ],
    },
    "4.0": {
        # From https://trustee.ietf.org/license-info/IETF-TLP-4.htm
        "draft": [
            """<t>
            This Internet-Draft is submitted in full conformance with the
            provisions of BCP 78 and BCP 79.
            </t>""",
        ],
        "IETF": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as the document
            authors. All rights reserved.
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents
            (http://trustee.ietf.org/license-info) in effect on the date of
            publication of this document. Please review these documents carefully,
            as they describe your rights and restrictions with respect to this
            document. Code Components extracted from this document must include
            Simplified BSD License text as described in Section 4.e of the Trust
            Legal Provisions and are provided without warranty as described in the
            Simplified BSD License.
            </t>""",
        ],
        "alt": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as the
            document authors. All rights reserved.
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents
            (http://trustee.ietf.org/license-info) in effect on the date of
            publication of this document. Please review these documents
            carefully, as they describe your rights and restrictions with
            respect to this document.
            </t>""",
        ],
        "noModification": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, except to format it for publication as an RFC or
            to translate it into languages other than English.
            </t>""",
        ],
        "noDerivatives": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, and it may not be published except as an
            Internet-Draft.
            </t>""",
        ],
        "pre5378": [
            """<t>
            This document may contain material from IETF Documents or IETF
            Contributions published or made publicly available before November
            10, 2008. The person(s) controlling the copyright in some of this
            material may not have granted the IETF Trust the right to allow
            modifications of such material outside the IETF Standards Process.
            Without obtaining an adequate license from the person(s)
            controlling the copyright in such materials, this document may not
            be modified outside the IETF Standards Process, and derivative
            works of it may not be created outside the IETF Standards Process,
            except to format it for publication as an RFC or to translate it
            into languages other than English.
            </t>""",
        ],
    },
    "5.0": {
        # From https://trustee.ietf.org/license-info/IETF-TLP-5.htm
        "draft": [
            """<t>
            This Internet-Draft is submitted in full conformance with the
            provisions of BCP 78 and BCP 79.
            </t>""",
        ],
        "IETF": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as the
            document authors. All rights reserved.
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents
            (<eref target="{scheme}://trustee.ietf.org/license-info"/>) in effect on the date of
            publication of this document. Please review these documents
            carefully, as they describe your rights and restrictions with
            respect to this document. Code Components extracted from this
            document must include Revised BSD License text as described in
            Section 4.e of the Trust Legal Provisions and are provided without
            warranty as described in the Revised BSD License.
            </t>""",
        ],
        "alt": [
            """<t>
            Copyright (c) {year} IETF Trust and the persons identified as the
            document authors. All rights reserved.
            </t>""",
            """<t>
            This document is subject to BCP 78 and the IETF Trust's Legal
            Provisions Relating to IETF Documents
            (<eref target="{scheme}://trustee.ietf.org/license-info"/>) in effect on the date of
            publication of this document. Please review these documents
            carefully, as they describe your rights and restrictions with
            respect to this document.
            </t>""",
        ],
        "noModification": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, except to format it for publication as an RFC or
            to translate it into languages other than English.
            </t>""",
        ],
        "noDerivatives": [
            """<t>
            This document may not be modified, and derivative works of it may
            not be created, and it may not be published except as an
            Internet-Draft.
            </t>""",
        ],
        "pre5378": [
            """<t>
            This document may contain material from IETF Documents or IETF
            Contributions published or made publicly available before November
            10, 2008. The person(s) controlling the copyright in some of this
            material may not have granted the IETF Trust the right to allow
            modifications of such material outside the IETF Standards Process.
            Without obtaining an adequate license from the person(s)
            controlling the copyright in such materials, this document may not
            be modified outside the IETF Standards Process, and derivative
            works of it may not be created outside the IETF Standards Process,
            except to format it for publication as an RFC or to translate it
            into languages other than English.
            </t>""",
        ],
    },
}
